// ============================================================================
// $Id: circle_protocol.cc,v 1.1 1996/04/22 18:25:24 questor Exp $
//
// This is part of Realm Of Magic
//
// Usage: Class that translates incoming messages for the circle protocol
//
// All rights reserved. See license.doc for complete information.
//
// Copyright (C) 1993-1996 by the Hackerz of the Realm Of Magic
// Realm Of Magic is based on CircleMUD, Copyright (C) 1993.
// CircleMUD is based on DikuMUD, Copyright (C) 1990, 1991.
// Original intermud_client.c (C) 1995, 96 by Chris Austin (Stryker@Tempus)
// Original lpc protocol (C) by Mark Lewis (Nostradamus@Zebedee)
//                           by J.W. Armstrong (Alvin@Sushi)
// ============================================================================

#include <string>
#include <list>
#include <stdio.h>
#include <fstream.h>
#include <arpa/inet.h>
#include <sys/un.h>

#include "circle_protocol.hh"
#include "imuddefs.hh"
#include "intermud.hh"
#include "tokenizer.hh"
#include "mud.hh"
#include "net.hh"
#include "utils.hh"
#include "log.hh"

extern _mud localmud;
extern _log interlog;
extern _net network;

void _circle::IncomingInterwiz(string &mess)
{
  tokenizer tokens(mess, "|");
  localmud << "inc-interwiz#@#" << tokens[1] << "#@#" << tokens[2]
           << "#@#" << tokens[3] << sendit;
  return;
}

void _circle::IncomingIntertell(string &mess)
{
  tokenizer tokens(mess, "|");
  localmud << "inc-tell#@#" << tokens[3] << "#@#" << tokens[4]
           << "#@#" << tokens[1] << "#@#" << tokens[5] << sendit;
  return;
}

void _circle::IncomingIntertellNR(string &mess)
{
  tokenizer tokens(mess, "|");
  localmud << "inc-tell-nr#@#" << tokens[3] << "#@#" << tokens[4]
           << "#@#" << tokens[1] << "#@#" << tokens[5] << sendit;
  return;
}

void _circle::IncomingInterwhoREQ(string &mess)
{
  tokenizer tokens(mess, "|");
  localmud << "inc-who-query#@#" << tokens[2] << "#@#" << tokens[3] << sendit;
  return;
}

void _circle::IncomingInterwho(string &mess)
{
  tokenizer tokens(mess, "|");
  if (!tokens[4])
    return;
  localmud << "inc-who-reply#@#" << tokens[2] << "#@#" << tokens[1]
           << "#@#";
  for (int i = 4; i < tokens.size(); i++)
    localmud << tokens[i] << "\n";
  localmud << sendit;
  return;
}

void _circle::IncomingInterpage(string &mess)
{
  tokenizer tokens(mess, "|");
  localmud << "inc-interpage#@#" << tokens[3] << "#@#" << tokens[4]
           << "#@#" << tokens[1] << "#@#" << tokens[5] << sendit;
  return;
}

void _circle::IncomingPingREP(string &mess, sockaddr_in &remote_address)
{
  tokenizer tokens(mess, "|");
  if (tokens[0] && tokens[3] && tokens[1] && tokens[2]) {
    
    list<dns_record>::iterator position;
    for (position = dns_table.begin(); position != dns_table.end();
        ++position)
      if ((*position).ip_address == remote_address.sin_addr.s_addr) {
        update_timestamp(*position);
        return;
      }
  } else {
    interlog << "INFO: Received ambiguous ping reply from "
             << inet_ntoa(remote_address.sin_addr) << logit;
  }
  return;
}

void _circle::IncomingPingREQ(string &mess, sockaddr_in &remote_address)
{
  tokenizer tokens(mess, "|");
  if (tokens[0] && tokens[1] && tokens[2] && tokens[3]) {
    list<dns_record>::iterator position;
    for (position = dns_table.begin(); position != dns_table.end();
        ++position)
      if ((*position).ip_address == remote_address.sin_addr.s_addr) {
        update_timestamp(*position);
        break;
      }
    network.SetAddr(remote_address.sin_addr.s_addr);
    network.SetPort(remote_address.sin_port);
    network << "1030|" << MUDNAME << "|" << MUDPORT << "|"
            << INTERMUD_PORT << "|" << MUD_TYPE << "|" << MUD_VERSION
            << "|" << INTERMUD_VERSION << "|" << encode_services(muddns)
            << "|" << sendit;
  }
  return;
}

void _circle::OutgoingInterwiz(string &player, string &data)
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
      ++position) 
    if (case_compare(muddns.mud_name, (*position).mud_name)) {
      network.SetAddr((*position).ip_address);
      network.SetPort((*position).port);
      network << "3000|" << player << "|" << MUDNAME << "|" << data
              << "|";
      update_statistics((*position).mud_name, 0, network.OutBuffer().size());
      network << sendit;
    }
  return;
}

bool _circle::OutgoingIntertell(string &recipent, string &remote_mud,
                                string &player, string &data)
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
      ++position)
    if (!case_compare((*position).mud_name, remote_mud)) {
      network.SetAddr((*position).ip_address);
      network.SetPort((*position).port);
      network << "2000|" << recipent << "|" << remote_mud << "|"
              << player << "|" << MUDNAME << "|" << data << "|";
      update_statistics((*position).mud_name, 0, network.OutBuffer().size());
      network << sendit;
      return true;
    }
  return false;
}

bool _circle::OutgoingIntertellNR(string &recipent, string &remote_mud,
                                  string &player, string &data)
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
      ++position)
    if (!case_compare((*position).mud_name, remote_mud)) {
      network.SetAddr((*position).ip_address);
      network.SetPort((*position).port);
      network << "2010|" << recipent << "|" << remote_mud << "|"
              << player << "|" << MUDNAME << "|" << data << "|";
      update_statistics((*position).mud_name, 0, network.OutBuffer().size());
      network << sendit;
      return true;
    }
  return false;
}

bool _circle::OutgoingInterwhoREQ(string &remote_mud, string &player)
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
      ++position)
    if (!case_compare((*position).mud_name, remote_mud)) {
      network.SetAddr((*position).ip_address);
      network.SetPort((*position).port);
      network << "2020|" << remote_mud << "|" << player << "|"
              << MUDNAME << "|";
      update_statistics((*position).mud_name, 0, network.OutBuffer().size());
      network << sendit;
      return true;
    }
  return false;
}

bool _circle::OutgoingInterwho(string &recipent, string &remote_mud,
                               string &data)
{
  tokenizer tokens(data, "\n");
  int total_data = 0;
  int count_lines = 0;
  int pak_count = 1;
  
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
      ++position)
    if (!case_compare((*position).mud_name, remote_mud)) {
      do {
        network << "2030|" << remote_mud << "|" << recipent << "|"
                << MUDNAME << "|";
        do {
          network << tokens[total_data] << "|";
          count_lines++;
          total_data++;
        } while (count_lines < 12 && total_data < tokens.size());
        update_statistics((*position).mud_name, 0, network.OutBuffer().size());
        network.SetAddr((*position).ip_address);
        network.SetPort((*position).port);
        network << sendit;
        count_lines = 0;
        pak_count++;
      } while (total_data < tokens.size());
      return true;
    }
  return false;
}
        
bool _circle::OutgoingInterpage(string &recipent, string &remote_mud,
                                string &player, string &data)
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
      ++position)
    if (!case_compare((*position).mud_name, remote_mud)) {
      network.SetAddr((*position).ip_address);
      network.SetPort((*position).port);
      network << "2040|" << recipent << "|" << remote_mud << "|"
              << player << "|" << MUDNAME << "|" << data << "|";
      update_statistics((*position).mud_name, 0, network.OutBuffer().size());
      network << sendit;
      return true;
    }
  return false;
}

void _circle::SendPings()
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
      ++position) 
    if (case_compare((*position).mud_name, MUDNAME)) {
      network.SetAddr((*position).ip_address);
      network.SetPort((*position).port);
      network << "1020|" << INTERMUD_PORT << "|" << MUDNAME << "|"
              << MUDPORT << "|" << MUD_TYPE << "|" << MUD_VERSION << "|"
              << INTERMUD_VERSION << "|" << encode_services(muddns)
              << "|" << sendit;
    }
  return;
}

void _circle::Startup()
{
  struct dns_record dummy;
  char buffer[8192];
  tokenizer tokens;
  
  if (dns_table.empty()) {
    muddns.ip_address = inet_addr(MUDIPADDR);
    muddns.port = htons(atoi(INTERMUD_PORT));
    muddns.mud_name = MUDNAME;
    muddns.mud_port = MUDPORT;
    muddns.mud_version = MUD_VERSION;
    muddns.intermud_version = INTERMUD_VERSION;
    muddns.mud_type = MUD_TYPE;
    muddns.kb_in = 0;
    muddns.kb_out = 0;
    muddns.msg_in = 0;
    muddns.msg_out = 0;
    muddns.muted = 0;
    muddns.last_connected = "Local Mud";
    if (SRV_INTERWIZ == 1)
      SET_BIT(muddns.services, ST_INTERWIZ);
    if (SRV_INTERTELL == 1)
      SET_BIT(muddns.services, ST_INTERTELL);
    if (SRV_INTERPAGE == 1)
      SET_BIT(muddns.services, ST_INTERPAGE);
    if (SRV_INTERWHO == 1)
      SET_BIT(muddns.services, ST_INTERWHO);
    dns_table.push_back(muddns);
    interlog << "INFO: Added DNS record for local mud" << logit;
  }
  ifstream mudlistfile(MUDLIST_PATH);
  if (!mudlistfile) {
    interlog << "SYSERR: No mudlist file, exiting" << logit;
    exit(1);
  }
  dummy.kb_in = 0;
  dummy.kb_out = 0;
  dummy.msg_in = 0;
  dummy.msg_out = 0;
  dummy.muted = 0;
  dummy.last_connected = "No connection yet";
  while (!mudlistfile.eof()) {
    mudlistfile.getline(buffer, 8192, '\n');
    tokens.tokenize(buffer, "|");
    if (tokens[0]) {
      dummy.ip_address = inet_addr(tokens[2]);
      dummy.port = htons(atoi(tokens[3]));
      dummy.mud_name = tokens[0];
      dummy.mud_port = tokens[1];
      dummy.mud_version = tokens[5];
      dummy.intermud_version = tokens[6];
      dummy.mud_type = atoi(tokens[4]);
      dummy.services = 0;
      decode_services(tokens[7], dummy);
      dns_table.push_back(dummy);
    }
    tokens.clear();
  }
  mudlistfile.close();
  return;
}

void _circle::PurgeList()
{
  struct dns_record temp;
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
      ++position) {
    if (!case_compare((*position).mud_name, MUDNAME))
      copy_dns_record(temp, *position);
  }
  dns_table.erase(dns_table.begin(), dns_table.end());
  dns_table.push_back(temp);
  return;
}

bool _circle::IncomingMessage(string &mess, sockaddr_in &remote_address)
{
  string msgnum_t;
  int msgnum;
  
  if (!isdigit(mess[0]))
    return false;

  msgnum_t = mess.substr(0, 4);
  msgnum = atoi(msgnum_t.c_str());

  switch(msgnum) {
    case 1020: IncomingPingREQ(mess, remote_address); break;
    case 1030: IncomingPingREP(mess, remote_address); break;
    case 2000: IncomingIntertell(mess); break;
    case 2010: IncomingIntertellNR(mess); break;
    case 2020: IncomingInterwhoREQ(mess); break;
    case 2030: IncomingInterwho(mess); break;
    case 2040: IncomingInterpage(mess); break;
    case 3000: IncomingInterwiz(mess); break;
    default:
      return false;
  }
  return true;
}
